from drf_yasg import openapi
from drf_yasg.views import get_schema_view
from rest_framework import permissions

info = openapi.Info(
    title="Notifications webhook receiver",
    default_version='v1',
    description="API Specification to be able to receive notifications from the NRC",
    contact=openapi.Contact(url="https://github.com/VNG-Realisatie/gemma-zaken"),
)


SchemaView = get_schema_view(
    public=True,
    permission_classes=(permissions.AllowAny,),
)



# Generated by Django 2.1.1 on 2019-03-19 09:48

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('notifications', '0002_subscription__subscription'),
    ]

    operations = [
        migrations.RenameField(
            model_name='notificationsconfig',
            old_name='location',
            new_name='api_root',
        ),
        migrations.AddField(
            model_name='notificationsconfig',
            name='client_id',
            field=models.CharField(blank=True, max_length=255, verbose_name='client id'),
        ),
        migrations.AddField(
            model_name='notificationsconfig',
            name='secret',
            field=models.CharField(blank=True, max_length=255, verbose_name='secret'),
        ),
        migrations.AlterField(
            model_name='notificationsconfig',
            name='api_root',
            field=models.URLField(unique=True, verbose_name='api root'),
        ),
    ]



import logging
from typing import Dict, List, Union
from urllib.parse import urlparse

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.utils import timezone

from djangorestframework_camel_case.util import camelize
from rest_framework.permissions import SAFE_METHODS
from rest_framework.routers import SimpleRouter
from zds_client import ClientError

from ..utils import get_resource_for_path
from .api.serializers import NotificatieSerializer
from .kanalen import Kanaal
from .models import NotificationsConfig

logger = logging.getLogger(__name__)


class NotificationMixinBase(type):

    def __new__(cls, name, bases, attrs):
        new_cls = super().__new__(cls, name, bases, attrs)

        kanaal = attrs.get('notifications_kanaal')
        if kanaal is None:
            return new_cls

        relevant_bases = [base for base in bases if issubclass(base, NotificationMixin)]
        resource = new_cls.queryset.model._meta.model_name
        # use the router to figure out which actions are available
        router = SimpleRouter()
        for route in router.get_routes(new_cls):
            for method, action in route.mapping.items():

                if method.upper() in SAFE_METHODS:
                    continue

                # check if the action is actually provided by one of the mixins
                if not any(hasattr(base, action) for base in relevant_bases):
                    continue

                kanaal.usage[resource].append(action)

        return new_cls


class NotificationMixin(metaclass=NotificationMixinBase):
    notifications_kanaal = None  # must be set be subclasses
    notifications_main_resource_key = None

    def get_kanaal(self):
        if not self.notifications_kanaal:
            raise ImproperlyConfigured(
                "'%s' should either include a `notifications_kanaal` "
                "attribute, or override the `get_kanaal()` method."
                % self.__class__.__name__
            )
        return self.notifications_kanaal

    def get_main_resource_key(self, kanaal: Kanaal) -> str:
        """
        Determine the key in the (response) data that represents the main resource.
        """
        if self.notifications_main_resource_key:
            return self.notifications_main_resource_key

        return kanaal.main_resource._meta.model_name

    def get_notification_main_object_url(self, data: dict, kanaal: Kanaal) -> str:
        """
        Retrieve the URL for the main object.
        """
        # using the main resource name, look up what the URL to this
        # object is/should be
        return data[self.get_main_resource_key(kanaal)]

    def construct_message(self, data: dict, instance: models.Model = None) -> dict:
        """
        Construct the message to send to the notification component.

        Using the response data from the view/action, we introspect this data
        to send the appropriate response. By convention, every resource
        includes its own absolute url in the 'url' key - we can use this to
        look up the object it points to. By convention, relations use the name
        of the resource, so for sub-resources we can use this to get a
        reference back to the main resource.
        """
        kanaal = self.get_kanaal()
        assert isinstance(kanaal, Kanaal), "`kanaal` should be a `Kanaal` instance"

        model = self.get_queryset().model

        if model is kanaal.main_resource:
            # look up the object in the database from its absolute URL
            resource_path = urlparse(data['url']).path
            resource = instance or get_resource_for_path(resource_path)

            main_object = resource
            main_object_url = data['url']
        else:
            # lookup the main object from the URL
            main_object_url = self.get_notification_main_object_url(data, kanaal)
            main_object_path = urlparse(main_object_url).path
            main_object = get_resource_for_path(main_object_path)

        message_data = {
            'kanaal': kanaal.label,
            'hoofd_object': main_object_url,
            'resource': model._meta.model_name,
            'resource_url': data['url'],
            'actie': self.action,
            'aanmaakdatum': timezone.now(),
            # each channel knows which kenmerken it has, so delegate this
            'kenmerken': kanaal.get_kenmerken(main_object),
        }

        # let the serializer & render machinery shape the data the way it
        # should be, suitable for JSON in/output
        serializer = NotificatieSerializer(instance=message_data)
        return camelize(serializer.data)

    def notify(self, status_code: int,
               data: Union[List, Dict], instance: models.Model = None) -> None:
        if settings.NOTIFICATIONS_DISABLED:
            return

        # do nothing unless we have a 'success' status code - early exit here
        if not 200 <= status_code < 300:
            logger.info("Not notifying, status code '%s' does not represent success.", status_code)
            return

        # build the content of the notification
        message = self.construct_message(data, instance=instance)

        # build the client from the singleton config. This will raise an
        # exception if the config is not complete. We want this to hard-fail!
        client = NotificationsConfig.get_client()
        try:
            client.create('notificaties', message)
        # any unexpected errors should show up in error-monitoring, so we only
        # catch ClientError exceptions
        except ClientError as exc:
            logger.warning("Could not deliver message to %s", client.base_url, exc_info=True)


class NotificationCreateMixin(NotificationMixin):
    def create(self, request, *args, **kwargs):
        response = super().create(request, *args, **kwargs)
        self.notify(response.status_code, response.data)
        return response


class NotificationUpdateMixin(NotificationMixin):
    def update(self, request, *args, **kwargs):
        response = super().update(request, *args, **kwargs)
        self.notify(response.status_code, response.data)
        return response


class NotificationDestroyMixin(NotificationMixin):
    def destroy(self, request, *args, **kwargs):
        # get data via serializer
        instance = self.get_object()
        data = self.get_serializer(instance).data

        response = super().destroy(request, *args, **kwargs)
        self.notify(response.status_code, data, instance=instance)
        return response


class NotificationViewSetMixin(NotificationCreateMixin,
                               NotificationUpdateMixin,
                               NotificationDestroyMixin):
    pass



# Generated by Django 2.0.13 on 2019-03-12 10:36

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('notifications', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='subscription',
            name='_subscription',
            field=models.URLField(blank=True, editable=False, help_text='Subscription as it is known in the NC', verbose_name='NC subscription'),
        ),
    ]



from django.views.generic import TemplateView

from .kanalen import KANAAL_REGISTRY


class KanalenView(TemplateView):
    template_name = 'vng_api_common/ref/kanalen.html'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['kanalen'] = sorted(KANAAL_REGISTRY, key=lambda s: s.label)
        return context



import os

from vng_api_common.conf.api import *  # noqa

DEBUG = os.getenv('DEBUG', 'no').lower() in ['yes', 'true', '1']

BASE_DIR = os.path.abspath(os.path.dirname(__file__))

SECRET_KEY = 'so-secret-i-cant-believe-you-are-looking-at-this'

ALLOWED_HOSTS = ['*']

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.postgresql',
        'NAME': 'vng_api_common',
        'USER': 'vng_api_common',
        'PASSWORD': 'vng_api_common',
    }
}

INSTALLED_APPS = [
    'django.contrib.contenttypes',
    'django.contrib.auth',
    'django.contrib.sessions',
    'django.contrib.admin',
    'django.contrib.staticfiles',
    'django.contrib.postgres',

    'rest_framework',
    'drf_yasg',

    'vng_api_common',
    'vng_api_common.authorizations',
    'vng_api_common.notifications',

    'testapp',
]

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

ROOT_URLCONF = 'testapp.urls'

STATIC_URL = '/static/'

REST_FRAMEWORK = BASE_REST_FRAMEWORK.copy()

SWAGGER_SETTINGS = BASE_SWAGGER_SETTINGS.copy()

SWAGGER_SETTINGS['DEFAULT_FIELD_INSPECTORS'] = SWAGGER_SETTINGS['DEFAULT_FIELD_INSPECTORS'][1:]

SWAGGER_SETTINGS.update({
    'DEFAULT_INFO': 'testapp.schema.info',
    'SECURITY_DEFINITIONS': {},
})



# Generated by Django 2.1.8 on 2019-05-02 04:09

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('authorizations', '0002_authorizationsconfig'),
    ]

    operations = [
        migrations.AlterField(
            model_name='authorizationsconfig',
            name='api_root',
            field=models.URLField(default='https://ref.tst.vng.cloud/ac/api/v1', unique=True, verbose_name='api root'),
        ),
    ]



import time

import jwt
from rest_framework import status

from ..authorizations.models import (
    Applicatie, AuthorizationsConfig, Autorisatie
)
from ..constants import VertrouwelijkheidsAanduiding
from ..models import JWTSecret


def generate_jwt(scopes: list, secret: str='letmein', zaaktypes: list=None) -> str:
    scope_labels = sum((_get_scope_labels(scope) for scope in scopes), [])
    payload = {
        # standard claims
        'iss': 'testsuite',
        'iat': int(time.time()),
        # custom claims
        'zds': {
            'scopes': scope_labels,
            'zaaktypes': zaaktypes or [],
        },
    }
    headers = {
        'client_identifier': 'testsuite',
    }
    encoded = jwt.encode(payload, secret, headers=headers, algorithm='HS256')
    encoded = encoded.decode('ascii')
    return f"Bearer {encoded}"


def _get_scope_labels(scope) -> list:
    if not scope.children:
        return [scope.label]

    labels = []
    for child in scope.children:
        labels += _get_scope_labels(child)
    return sorted(set(labels))


class AuthCheckMixin:

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        JWTSecret.objects.get_or_create(
            identifier='testsuite',
            defaults={'secret': 'letmein'}
        )

    def assertForbidden(self, url, method='get', request_kwargs=None):
        """
        Assert that an appropriate scope is required.
        """
        do_request = getattr(self.client, method)
        request_kwargs = request_kwargs or {}

        with self.subTest(case='JWT missing'):
            response = do_request(url, **request_kwargs)

            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        with self.subTest(case='Invalid JWT structure'):
            invalid_jwt = generate_jwt_auth('testsuite', 'letmein')[:-10]
            self.client.credentials(HTTP_AUTHORIZATION=invalid_jwt)

            response = do_request(url, **request_kwargs)

            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def assertForbiddenWithCorrectScope(
            self, url: str, scopes: list, method='get',
            request_kwargs=None, **extra_claims):

        do_request = getattr(self.client, method)
        request_kwargs = request_kwargs or {}

        jwt = generate_jwt(scopes=scopes, **extra_claims)
        self.client.credentials(HTTP_AUTHORIZATION=jwt)

        response = do_request(url, **request_kwargs)

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)


class JWTScopesMixin:

    scopes = None
    zaaktypes = None

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        JWTSecret.objects.get_or_create(
            identifier='testsuite',
            defaults={'secret': 'letmein'}
        )

    def setUp(self):
        super().setUp()

        if self.scopes is not None:
            token = generate_jwt(
                scopes=self.scopes,
                zaaktypes=self.zaaktypes or [],
                secret='letmein'
            )
            self.client.credentials(HTTP_AUTHORIZATION=token)


# tools fot testing with new authorization format
def generate_jwt_auth(client_id, secret):
    payload = {
        # standard claims
        'iss': 'testsuite',
        'iat': int(time.time()),
        # custom
        'client_id': client_id
    }
    encoded = jwt.encode(payload, secret, algorithm='HS256')
    encoded = encoded.decode('ascii')
    return f"Bearer {encoded}"


class JWTAuthMixin:
    scopes = None
    heeft_alle_autorisaties = False
    zaaktype = None
    max_vertrouwelijkheidaanduiding = VertrouwelijkheidsAanduiding.zeer_geheim

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        JWTSecret.objects.get_or_create(
            identifier='testsuite',
            defaults={'secret': 'letmein'}
        )

        config = AuthorizationsConfig.get_solo()

        cls.applicatie = Applicatie.objects.create(
            client_ids=['testsuite'],
            label='for test',
            heeft_alle_autorisaties=cls.heeft_alle_autorisaties
        )

        if cls.heeft_alle_autorisaties is False:
            cls.autorisatie = Autorisatie.objects.create(
                applicatie=cls.applicatie,
                component=config.component,
                scopes=cls.scopes or [],
                zaaktype=cls.zaaktype or '',
                max_vertrouwelijkheidaanduiding=cls.max_vertrouwelijkheidaanduiding
            )

    def setUp(self):
        super().setUp()

        token = generate_jwt_auth(client_id='testsuite', secret='letmein')
        self.client.credentials(HTTP_AUTHORIZATION=token)



from django.contrib import admin

from solo.admin import SingletonModelAdmin

from .models import Applicatie, AuthorizationsConfig, Autorisatie


@admin.register(AuthorizationsConfig)
class AuthorizationsConfigAdmin(SingletonModelAdmin):
    list_display = ('api_root', 'component',)


@admin.register(Autorisatie)
class AutorisatieAdmin(admin.ModelAdmin):
    list_display = ('applicatie', 'component', 'zaaktype', 'scopes', 'max_vertrouwelijkheidaanduiding')


class AutorisatieInline(admin.TabularInline):
    model = Autorisatie


@admin.register(Applicatie)
class ApplicatieAdmin(admin.ModelAdmin):
    list_display = ('uuid', 'client_ids', 'label', 'heeft_alle_autorisaties', )
    readonly_fields = ('uuid',)
    inlines = (AutorisatieInline,)



from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _


class NotificationsConfig(AppConfig):
    name = 'vng_api_common.notifications'
    verbose_name = _("Notificaties")



from typing import Union
from urllib.parse import urlparse

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

from rest_framework import permissions
from rest_framework.renderers import BrowsableAPIRenderer
from rest_framework.request import Request

from .scopes import Scope
from .utils import get_resource_for_path


class ActionScopesRequired(permissions.BasePermission):
    """
    Look at the scopes required for the current action and check that they
    are present in the JWT
    """

    def get_required_scopes(self, view) -> Union[Scope, None]:
        if not hasattr(view, 'required_scopes'):
            raise ImproperlyConfigured("The View(Set) must have a `required_scopes` attribute")

        scopes_required = view.required_scopes.get(view.action)
        return scopes_required

    def has_permission(self, request: Request, view) -> bool:
        if settings.DEBUG and isinstance(request.accepted_renderer, BrowsableAPIRenderer):
            return True

        scopes_needed = self.get_required_scopes(view)
        # TODO: if no scopes are needed, what do???
        return request.jwt_payload.has_scopes(scopes_needed)


class ScopesRequired(permissions.BasePermission):
    """
    Very simple scope-based permission, does not map to HTTP method or action.
    """

    def has_permission(self, request: Request, view) -> bool:
        # don't enforce them in the browsable API during debugging/development
        if settings.DEBUG and isinstance(request.accepted_renderer, BrowsableAPIRenderer):
            return True

        if not hasattr(view, 'required_scopes'):
            raise ImproperlyConfigured("The View(Set) must have a `required_scopes` attribute")

        return request.jwt_payload.has_scopes(view.required_scopes)


class ClientIdRequired(permissions.BasePermission):
    """
    Look at the client_id of an object and check that it equals client_id in the JWT
    """

    def has_object_permission(self, request: Request, view, obj) -> bool:
        if settings.DEBUG and isinstance(request.accepted_renderer, BrowsableAPIRenderer):
            return True

        if request.method in permissions.SAFE_METHODS:
            return True
        else:
            return request.jwt_payload['client_id'] == obj.client_id


class AuthScopesRequired(permissions.BasePermission):
    """
    Look at the scopes required for the current action
    and check that they are present in the AC for this client
    """

    def get_required_scopes(self, view) -> Union[Scope, None]:
        if not hasattr(view, 'required_scopes'):
            raise ImproperlyConfigured("The View(Set) must have a `required_scopes` attribute")

        scopes_required = view.required_scopes.get(view.action)
        return scopes_required

    def get_zaaktype(self, obj):
        return None

    def get_vertrouwelijkheidaanduiding(self, obj):
        return None

    def get_zaaktype_from_request(self, request):
        return None

    def get_vertrouwelijkheidaanduiding_from_request(self, request):
        return None

    def has_permission(self, request: Request, view) -> bool:
        if settings.DEBUG and isinstance(request.accepted_renderer, BrowsableAPIRenderer):
            return True

        scopes_required = self.get_required_scopes(view)
        if view.action == 'create':
            zaaktype = self.get_zaaktype_from_request(request)
            vertrouwelijkheidaanduiding = self.get_vertrouwelijkheidaanduiding_from_request(request)
            return request.jwt_auth.has_auth(scopes_required, zaaktype, vertrouwelijkheidaanduiding)
        elif view.action == 'list':
            return request.jwt_auth.has_auth(scopes_required, None, None)

        return True

    def has_object_permission(self, request: Request, view, obj) -> bool:
        if settings.DEBUG and isinstance(request.accepted_renderer, BrowsableAPIRenderer):
            return True

        scopes_required = self.get_required_scopes(view)
        zaaktype = self.get_zaaktype(obj)
        vertrouwelijkheidaanduiding = self.get_vertrouwelijkheidaanduiding(obj)
        return request.jwt_auth.has_auth(scopes_required, zaaktype, vertrouwelijkheidaanduiding)


# TODO: subclasses should be in their respective components?


class ZaakAuthScopesRequired(AuthScopesRequired):
    """
    Look at the scopes required for the current action and at zaaktype and vertrouwelijkheidaanduiding
    of current zaak and check that they are present in the AC for this client
    """

    def get_zaaktype(self, obj):
        return obj.zaaktype

    def get_vertrouwelijkheidaanduiding(self, obj):
        return obj.vertrouwelijkheidaanduiding

    def get_zaaktype_from_request(self, request):
        return request.data.get('zaaktype', None)

    def get_vertrouwelijkheidaanduiding_from_request(self, request):
        return request.data.get('vertrouwelijkheidaanduiding', None)


class ZaakRelatedAuthScopesRequired(AuthScopesRequired):
    """
    Look at the scopes required for the current action and at zaaktype and vertrouwelijkheidaanduiding
    of related zaak and check that they are present in the AC for this client
    """

    def get_zaaktype(self, obj):
        return obj.zaak.zaaktype

    def get_vertrouwelijkheidaanduiding(self, obj):
        return obj.zaak.vertrouwelijkheidaanduiding

    def get_zaaktype_from_request(self, request):
        zaak_url = urlparse(request.data['zaak']).path
        zaak = get_resource_for_path(zaak_url)
        return zaak.zaaktype

    def get_vertrouwelijkheidaanduiding_from_request(self, request):
        zaak_url = urlparse(request.data['zaak']).path
        zaak = get_resource_for_path(zaak_url)
        return zaak.vertrouwelijkheidaanduiding



# Generated by Django 2.0.13 on 2019-04-17 11:45

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('vng_api_common', '0002_apicredential'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='apicredential',
            options={'verbose_name': 'external API credential', 'verbose_name_plural': 'external API credentials'},
        ),
        migrations.AlterModelOptions(
            name='jwtsecret',
            options={'verbose_name': 'client credential', 'verbose_name_plural': 'client credentials'},
        ),
        migrations.AddField(
            model_name='apicredential',
            name='label',
            field=models.CharField(default='', help_text='human readable label', max_length=100, verbose_name='label'),
        ),
    ]



from typing import Any

from django.db.models.base import ModelBase


def field_default(field: str, default: Any):

    def decorator(cls: ModelBase):
        model_field = cls._meta.get_field(field)
        model_field.default = default
        return cls

    return decorator



import logging

from django.db import transaction
from django.utils.translation import ugettext_lazy as _

from rest_framework import serializers

from vng_api_common.authorizations.models import Applicatie, Autorisatie

logger = logging.getLogger(__name__)


class AutorisatieSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = Autorisatie
        fields = (
            'component',
            'scopes',
            'zaaktype',
            'max_vertrouwelijkheidaanduiding',
        )


class ApplicatieSerializer(serializers.HyperlinkedModelSerializer):
    autorisaties = AutorisatieSerializer(many=True, required=False)

    class Meta:
        model = Applicatie
        fields = (
            'url',
            'client_ids',
            'label',
            'heeft_alle_autorisaties',
            'autorisaties'
        )
        extra_kwargs = {
            'url': {
                'lookup_field': 'uuid',
            },
            'heeft_alle_autorisaties': {
                'required': False,
            }

        }

    def validate(self, attrs):
        validated_attrs = super().validate(attrs)

        # either autorisaties or heeft_alle_autorisaties can be specified
        autorisaties_obj = None
        heeft_alle_autorisaties_obj = None
        # in case of update:
        if self.instance:
            autorisaties_obj = self.instance.autorisaties.all()
            heeft_alle_autorisaties_obj = self.instance.heeft_alle_autorisaties

        autorisaties = validated_attrs.get('autorisaties', autorisaties_obj)
        heeft_alle_autorisaties = validated_attrs.get('heeft_alle_autorisaties', heeft_alle_autorisaties_obj)

        if autorisaties and heeft_alle_autorisaties is True:
            raise serializers.ValidationError(
                _('Either autorisaties or heeft_alle_autorisaties can be specified'),
                code='ambiguous-authorizations-specified')

        if not autorisaties and heeft_alle_autorisaties is not True:
            raise serializers.ValidationError(
                _('Either autorisaties or heeft_alle_autorisaties should be specified'),
                code='missing-authorizations')

        return validated_attrs

    @transaction.atomic
    def create(self, validated_data):
        autorisaties_data = validated_data.pop('autorisaties', None)
        applicatie = super().create(validated_data)

        if autorisaties_data:
            for auth in autorisaties_data:
                Autorisatie.objects.create(**auth, applicatie=applicatie)

        return applicatie

    @transaction.atomic
    def update(self, instance, validated_data):
        autorisaties_data = validated_data.pop('autorisaties', None)
        applicatie = super().update(instance, validated_data)

        # in case of update autorisaties - remove all related autorisaties
        if autorisaties_data:
            applicatie.autorisaties.all().delete()
            for auth in autorisaties_data:
                Autorisatie.objects.create(**auth, applicatie=applicatie)

        return applicatie


class ApplicatieUuidSerializer(ApplicatieSerializer):
    """
    Serializer for saving data in local auth DB
    uuid is used for synchronizing identifiers with AC DB
    """
    class Meta(ApplicatieSerializer.Meta):
        fields = ApplicatieSerializer.Meta.fields + ('uuid', )



from django.contrib import admin
from django.urls import include, path, re_path

from .schema import SchemaView
from .views import NotificationView

urlpatterns = [
    path('admin/', admin.site.urls),

    path('api/', include([
        # API documentation
        re_path(r'^schema/openapi(?P<format>\.json|\.yaml)$',
                SchemaView.without_ui(cache_timeout=None),
                name='schema-json'),
        re_path(r'^schema/$',
                SchemaView.with_ui('redoc', cache_timeout=None),
                name='schema-redoc'),

    ])),

    # this is a hack to get the parameter to show up in the API spec
    # this effectively makes this a wildcard URL, so it should be LAST
    path('<webhooks_path>', NotificationView.as_view()),
]



# https://pyjwt.readthedocs.io/en/latest/usage.html#reading-headers-without-validation
# -> we can put the organization/service in the headers itself
import logging
from typing import List, Union

from django.conf import settings
from django.db import models, transaction
from django.db.models import Q
from django.utils.functional import cached_property
from django.utils.translation import ugettext as _

import jwt
from djangorestframework_camel_case.util import underscoreize
from rest_framework.exceptions import PermissionDenied
from rest_framework.response import Response
from zds_client.client import ClientError

from .authorizations.models import (
    Applicatie, AuthorizationsConfig, Autorisatie
)
from .authorizations.serializers import ApplicatieUuidSerializer
from .constants import VERSION_HEADER
from .models import JWTSecret
from .scopes import Scope
from .utils import get_uuid_from_path

logger = logging.getLogger(__name__)

EMPTY_PAYLOAD = {
    'scopes': [],
    'zaaktypes': [],
}


class JWTPayload:

    def __init__(self, encoded: str=None):
        self.encoded = encoded

    def __repr__(self):
        return "<%s: payload=%r>" % (self.__class__.__name__, self.payload)

    def __getitem__(self, key):
        return self.payload[key]

    def get(self, attr, *args, **kwargs):
        return self.payload.get(attr, *args, **kwargs)

    @cached_property
    def payload(self) -> dict:
        if self.encoded is None:
            return EMPTY_PAYLOAD

        try:
            header = jwt.get_unverified_header(self.encoded)
        except jwt.DecodeError:
            logger.info("Invalid JWT encountered")
            raise PermissionDenied(
                _('JWT could not be decoded. Possibly you made a copy-paste mistake.'),
                code='jwt-decode-error'
            )

        try:
            jwt_secret = JWTSecret.objects.get(identifier=header['client_identifier'])
        except JWTSecret.DoesNotExist:
            raise PermissionDenied(
                'Client identifier bestaat niet',
                code='invalid-client-identifier'
            )
        except KeyError:
            raise PermissionDenied(
                'Client identifier is niet aanwezig in JWT',
                code='missing-client-identifier'
            )
        else:
            key = jwt_secret.secret

        # the jwt package does verification against tampering (TODO: unit test)
        try:
            payload = jwt.decode(self.encoded, key, algorithms='HS256')
        except jwt.InvalidSignatureError as exc:
            logger.exception("Invalid signature - possible payload tampering?")
            raise PermissionDenied(
                'Client credentials zijn niet geldig',
                code='invalid-jwt-signature'
            )

        res = payload.get('zds', EMPTY_PAYLOAD)
        res['client_id'] = header['client_identifier']

        return res

    def has_scopes(self, scopes: Union[Scope, None]) -> bool:
        """
        Check whether all of the expected scopes are present or not.
        """
        if scopes is None:
            # TODO: should block instead of allow it - we'll gradually introduce this
            return True

        scopes_provided = self.payload.get('scopes', [])
        logger.debug("Scopes provided are: %s", scopes)
        # simple form - needs a more complex setup if a scope 'bundles'
        # other scopes
        return scopes.is_contained_in(scopes_provided)


class JWTAuth:
    def __init__(self, encoded: str = None):
        self.encoded = encoded

    @property
    def applicaties(self) -> Union[list, None]:
        if self.client_id is None:
            return []

        applicaties = self._get_auth()

        if not applicaties:
            auth_data = self._request_auth()
            applicaties = self._save_auth(auth_data)

        return applicaties

    @property
    def autorisaties(self) -> models.QuerySet:
        """
        Retrieve all authorizations relevant to this component.
        """
        app_ids = [app.id for app in self.applicaties]
        config = AuthorizationsConfig.get_solo()
        return (
            Autorisatie.objects
            .filter(
                applicatie_id__in=app_ids,
                component=config.component
            )
        )

    def _request_auth(self) -> list:
        client = AuthorizationsConfig.get_client()
        try:
            response = client.list(
                'applicatie',
                query_params={'client_ids': self.client_id}
            )
        except ClientError:
            logger.warn("Authorization component can't be accessed")
            return []

        return underscoreize(response['results'])

    def _get_auth(self):
        return Applicatie.objects.filter(client_ids__contains=[self.client_id])

    @transaction.atomic
    def _save_auth(self, auth_data):
        applicaties = []

        for applicatie_data in auth_data:
            applicatie_serializer = ApplicatieUuidSerializer(data=applicatie_data)
            uuid = get_uuid_from_path(applicatie_data['url'])
            applicatie_data['uuid'] = uuid
            applicatie_serializer.is_valid()
            applicaties.append(applicatie_serializer.save())

        return applicaties

    @cached_property
    def client_id(self) -> Union[str, None]:
        if self.encoded is None:
            return None

        # jwt check
        try:
            payload = jwt.decode(self.encoded, verify=False)
        except jwt.DecodeError:
            logger.info("Invalid JWT encountered")
            raise PermissionDenied(
                _('JWT could not be decoded. Possibly you made a copy-paste mistake.'),
                code='jwt-decode-error'
            )

        # get client_id
        try:
            client_id = payload['client_id']
        except KeyError:
            try:
                header = jwt.get_unverified_header(self.encoded)
            except jwt.DecodeError:
                logger.info("Invalid JWT encountered")
                raise PermissionDenied(
                    _('JWT could not be decoded. Possibly you made a copy-paste mistake.'),
                    code='jwt-decode-error'
                )
            else:
                try:
                    client_id = header['client_identifier']
                except KeyError:
                    raise PermissionDenied(
                        'Client identifier is niet aanwezig in JWT',
                        code='missing-client-identifier'
                    )
                else:
                    logger.warning(_('The support of authorization old format will be terminated soon. '
                                     'Please use a new format with the separate Authorization Component'))

        # find client_id in DB and retrieve it's secret
        try:
            jwt_secret = JWTSecret.objects.get(identifier=client_id)
        except JWTSecret.DoesNotExist:
            raise PermissionDenied(
                'Client identifier bestaat niet',
                code='invalid-client-identifier'
            )
        else:
            key = jwt_secret.secret

        # check signature of the token
        try:
            jwt.decode(self.encoded, key, algorithms='HS256')
        except jwt.InvalidSignatureError as exc:
            logger.exception("Invalid signature - possible payload tampering?")
            raise PermissionDenied(
                'Client credentials zijn niet geldig',
                code='invalid-jwt-signature'
            )

        return client_id

    def has_auth(self, scopes: List[str], zaaktype: Union[str, None],
                 vertrouwelijkheidaanduiding: Union[str, None]) -> bool:

        if scopes is None:
            return False

        scopes_provided = set()
        config = AuthorizationsConfig.get_solo()

        for applicatie in self.applicaties:
            # allow everything
            if applicatie.heeft_alle_autorisaties is True:
                return True

            # consider all scopes at all zaaktypes and vertrouwelijkheidaanduiding
            zaaktype_q = Q(zaaktype=zaaktype) if zaaktype is not None else Q()

            for autorisatie in applicatie.autorisaties.filter(zaaktype_q, component=config.component):
                vertrouwelijkheidaanduiding_ok = (
                    vertrouwelijkheidaanduiding is None
                    or autorisatie.satisfy_vertrouwelijkheid(vertrouwelijkheidaanduiding)  # noqa
                )

                if vertrouwelijkheidaanduiding_ok:
                    scopes_provided.update(autorisatie.scopes)

        return scopes.is_contained_in(list(scopes_provided))


class AuthMiddleware:

    header = 'HTTP_AUTHORIZATION'
    auth_type = 'Bearer'

    def __init__(self, get_response=None):
        self.get_response = get_response

    def __call__(self, request):
        self.extract_jwt_payload(request)
        return self.get_response(request) if self.get_response else None

    def extract_jwt_payload(self, request):
        authorization = request.META.get(self.header, '')
        prefix = f"{self.auth_type} "
        if authorization.startswith(prefix):
            # grab the actual token
            encoded = authorization[len(prefix):]
        else:
            encoded = None

        request.jwt_payload = JWTPayload(encoded)
        request.jwt_auth = JWTAuth(encoded)


class APIVersionHeaderMiddleware:
    """
    Include a header specifying the API-version
    """

    def __init__(self, get_response=None):
        self.get_response = get_response

    def __call__(self, request):
        if self.get_response is None:
            return None

        response = self.get_response(request)

        # not an API response, exit early
        if not isinstance(response, Response):
            return response

        # set the header
        response[VERSION_HEADER] = settings.API_VERSION

        return response



# Generated by Django 2.0.13 on 2019-04-17 11:42

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('notifications', '0005_fix_default_nrc'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='notificationsconfig',
            name='client_id',
        ),
        migrations.RemoveField(
            model_name='notificationsconfig',
            name='secret',
        ),
        migrations.AlterField(
            model_name='notificationsconfig',
            name='api_root',
            field=models.URLField(default='https://ref.tst.vng.cloud/nrc/api/v1', unique=True, verbose_name='api root'),
        ),
    ]



from django.utils.translation import ugettext_lazy as _

from rest_framework import serializers


class NotificatieSerializer(serializers.Serializer):
    kanaal = serializers.CharField(label=_("kanaal"), max_length=50)
    hoofd_object = serializers.URLField(label=_("URL naar het hoofdobject"))
    resource = serializers.CharField(label=_("resource"), max_length=100)
    resource_url = serializers.URLField(label=_("URL naar de resource waarover de notificatie gaat"))
    actie = serializers.CharField(label=_("actie"), max_length=100)
    aanmaakdatum = serializers.DateTimeField(label=_("aanmaakdatum"))
    kenmerken = serializers.DictField(
        child=serializers.CharField(max_length=1000),
        required=False
    )



import uuid

from django.contrib.postgres.fields import ArrayField
from django.db import models
from django.utils.translation import ugettext_lazy as _

from ..constants import ComponentTypes, VertrouwelijkheidsAanduiding
from ..decorators import field_default
from ..fields import VertrouwelijkheidsAanduidingField
from ..models import APIMixin, ClientConfig


@field_default('api_root', 'https://ref.tst.vng.cloud/ac/api/v1')
class AuthorizationsConfig(ClientConfig):
    component = models.CharField(
        _("component"), max_length=50, default=ComponentTypes.zrc,
        choices=ComponentTypes.choices
    )

    class Meta:
        verbose_name = _("Autorisatiecomponentconfiguratie")


class Applicatie(APIMixin, models.Model):
    """
    Client level of authorization
    """
    uuid = models.UUIDField(
        unique=True, default=uuid.uuid4,
        help_text="Unique resource identifier (UUID4)"
    )
    client_ids = ArrayField(
        models.CharField(max_length=50),
        help_text=_("Comma-separated list of identifiers used for authentication")
    )
    label = models.CharField(
        max_length=100,
        help_text=_("A human readable representation of the application")
    )
    heeft_alle_autorisaties = models.BooleanField(
        default=False,
        help_text=_("Globally allows everything")
    )

    def __str__(self):
        return f'Applicatie ({self.label})'


class Autorisatie(APIMixin, models.Model):
    applicatie = models.ForeignKey(
        'Applicatie',
        on_delete=models.CASCADE,
        related_name='autorisaties'
    )
    component = models.CharField(
        max_length=50,
        help_text=_("Name of the component to authorize")
    )
    zaaktype = models.URLField(
        help_text="Url of the zaaktype that is allowed",
        max_length=1000, blank=True
    )
    scopes = ArrayField(
        models.CharField(max_length=100),
        help_text=_("Comma-separated list of identifiers used for authentication")
    )
    max_vertrouwelijkheidaanduiding = VertrouwelijkheidsAanduidingField(
        help_text=_("Maximum level of confidentiality that is allowed"),
        blank=True
    )

    def satisfy_vertrouwelijkheid(self, vertrouwelijkheidaanduiding: str) -> bool:
        max_confid_level = VertrouwelijkheidsAanduiding.get_choice(self.max_vertrouwelijkheidaanduiding).order
        provided_confid_level = VertrouwelijkheidsAanduiding.get_choice(vertrouwelijkheidaanduiding).order
        return max_confid_level >= provided_confid_level



from drf_yasg.app_settings import swagger_settings

from vng_api_common.notifications.api.views import (
    NotificationView as _NotificationView
)


class NotificationView(_NotificationView):
    """
    Ontvang notificaties via webhook
    """
    swagger_schema = swagger_settings.DEFAULT_AUTO_SCHEMA_CLASS



# Generated by Django 2.1.8 on 2019-05-02 04:04

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('authorizations', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='AuthorizationsConfig',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('api_root', models.URLField(unique=True, verbose_name='api root')),
                ('component', models.CharField(choices=[('ZRC', 'ZRC'), ('ZTC', 'ZTC'), ('DRC', 'DRC'), ('BRC', 'BRC'), ('NC', 'NC'), ('AC', 'AC')], default='ZRC', max_length=50, verbose_name='component')),
            ],
            options={
                'verbose_name': 'Autorisatiecomponentconfiguratie',
            },
        ),
    ]



# Generated by Django 2.1.8 on 2019-04-23 09:32

import uuid

import django.contrib.postgres.fields
import django.db.models.deletion
from django.db import migrations, models

import vng_api_common.fields
import vng_api_common.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Applicatie',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('uuid', models.UUIDField(default=uuid.uuid4, help_text='Unique resource identifier (UUID4)', unique=True)),
                ('client_ids', django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=50), help_text='Comma-separated list of identifiers used for authentication', size=None)),
                ('label', models.CharField(help_text='A human readable representation of the application', max_length=100)),
                ('heeft_alle_autorisaties', models.BooleanField(default=False, help_text='Globally allows everything')),
            ],
            bases=(vng_api_common.models.APIMixin, models.Model),
        ),
        migrations.CreateModel(
            name='Autorisatie',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('component', models.CharField(help_text='Name of the component to authorize', max_length=50)),
                ('zaaktype', models.URLField(help_text='Url of the zaaktype that is allowed', max_length=1000)),
                ('scopes', django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=100), help_text='Comma-separated list of identifiers used for authentication', size=None)),
                ('max_vertrouwelijkheidaanduiding', vng_api_common.fields.VertrouwelijkheidsAanduidingField(choices=[('openbaar', 'OPENBAAR'), ('beperkt openbaar', 'BEPERKT OPENBAAR'), ('intern', 'INTERN'), ('zaakvertrouwelijk', 'ZAAKVERTROUWELIJK'), ('vertrouwelijk', 'VERTROUWELIJK'), ('confidentieel', 'CONFIDENTIEEL'), ('geheim', 'GEHEIM'), ('zeer geheim', 'ZEER GEHEIM')], help_text='Maximum level of confidentiality that is allowed', max_length=20)),
                ('applicatie', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='autorisaties', to='authorizations.Applicatie')),
            ],
            bases=(vng_api_common.models.APIMixin, models.Model),
        ),
    ]



import logging

from drf_yasg import openapi
from drf_yasg.inspectors.base import NotHandled
from drf_yasg.inspectors.field import FieldInspector
from rest_framework import serializers

logger = logging.getLogger(__name__)


TYPES_MAP = {
    str: openapi.TYPE_STRING,
    int: openapi.TYPE_INTEGER,
    bool: openapi.TYPE_BOOLEAN,
}


class ReadOnlyFieldInspector(FieldInspector):
    """
    Provides conversion for derived ReadOnlyField from model fields.

    This inspector looks at the type hint to determine the type/format of
    a model property.
    """

    def field_to_swagger_object(self, field, swagger_object_type, use_references, **kwargs):
        SwaggerType, ChildSwaggerType = self._get_partial_types(field, swagger_object_type, use_references, **kwargs)

        if isinstance(field, serializers.ReadOnlyField) and swagger_object_type == openapi.Schema:
            prop = getattr(field.parent.Meta.model, field.source)
            if not isinstance(prop, property):
                return NotHandled

            return_type = prop.fget.__annotations__.get('return')
            if return_type is None:  # no type annotation, too bad...
                logger.debug("Missing return type annotation for prop %s on model %s",
                             field.source, field.parent.Meta.model)
                return NotHandled

            type_ = TYPES_MAP.get(return_type)
            if type_ is None:
                logger.debug("Missing type mapping for %r", return_type)

            return SwaggerType(type=type_ or openapi.TYPE_STRING)

        return NotHandled



import logging

from django.conf import settings
from django.utils.module_loading import import_string

from djangorestframework_camel_case.util import underscoreize

from ..authorizations.models import Applicatie
from ..authorizations.serializers import ApplicatieUuidSerializer
from ..utils import get_uuid_from_path


class LoggingHandler:

    def handle(self, message: dict) -> None:
        logger = logging.getLogger('notifications')
        logger.info("Received notification %r", message)


class AuthHandler:
    
    def _request_auth(self, path: str) -> dict:
        Client = import_string(settings.ZDS_CLIENT_CLASS)
        client = Client.from_url(path)
        response = client.retrieve('applicatie', path)
        return underscoreize(response)

    def handle(self, message: dict) -> None:
        uuid = get_uuid_from_path(message['resource_url'])

        if message['actie'] == 'delete':
            Applicatie.objects.filter(uuid=uuid).delete()
            return

        # get info
        applicatie_data = self._request_auth(message['resource_url'])
        applicatie_data['uuid'] = uuid

        # update models
        try:
            applicatie = Applicatie.objects.get(uuid=uuid)
        except Applicatie.DoesNotExist:
            applicatie_serializer = ApplicatieUuidSerializer(data=applicatie_data)
        else:
            applicatie_serializer = ApplicatieUuidSerializer(applicatie, data=applicatie_data)
        applicatie_serializer.is_valid()
        applicatie_serializer.save()


class RoutingHandler:

    def __init__(self, config: dict, default=None):
        self.config = config
        self.default = default

    def handle(self, message: dict):
        handler = self.config.get(message['kanaal'])
        if handler is not None:
            handler.handle(message)
        elif self.default:
            self.default.handle(message)


log = LoggingHandler()
auth = AuthHandler()

default = RoutingHandler({'autorisaties': auth}, default=log)



# Generated by Django 2.0.13 on 2019-03-25 13:13

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('notifications', '0003_auto_20190319_1048'),
    ]

    operations = [
        migrations.AlterField(
            model_name='notificationsconfig',
            name='api_root',
            field=models.URLField(default='https://ref.tst.vng.cloud/nc/api/v1', unique=True, verbose_name='api root'),
        ),
    ]



# Generated by Django 2.0.13 on 2019-03-12 10:23

import django.contrib.postgres.fields
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='NotificationsConfig',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('location', models.URLField(help_text='API Root of the NC to use', verbose_name='location')),
            ],
            options={
                'verbose_name': 'Notificatiescomponentconfiguratie',
            },
        ),
        migrations.CreateModel(
            name='Subscription',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('callback_url', models.URLField(help_text='Where to send the notifications (webhook url)', verbose_name='callback url')),
                ('client_id', models.CharField(help_text='Client ID to construct the auth token', max_length=50, verbose_name='client ID')),
                ('secret', models.CharField(help_text='Secret to construct the auth token', max_length=50, verbose_name='client secret')),
                ('channels', django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=100), help_text='Comma-separated list of channels to subscribe to', size=None, verbose_name='channels')),
                ('config', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='notifications.NotificationsConfig')),
            ],
            options={
                'verbose_name': 'Webhook subscription',
                'verbose_name_plural': 'Webhook subscriptions',
            },
        ),
    ]
